@@ -1,5 +1,5 @@
 /* Variable expansion functions for GNU Make.
-Copyright (C) 1988, 1989, 1991, 1992, 1993 Free Software Foundation, Inc.
+Copyright (C) 1988, 89, 91, 92, 93, 95 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -14,21 +14,34 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU Make; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
+the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
+
+#include <assert.h>
 
 #include "make.h"
+#include "filedef.h"
+#include "job.h"
 #include "commands.h"
-#include "file.h"
 #include "variable.h"
+#include "rule.h"
 
 /* The next two describe the variable output buffer.
    This buffer is used to hold the variable-expansion of a line of the
    makefile.  It is made bigger with realloc whenever it is too small.
    variable_buffer_length is the size currently allocated.
-   variable_buffer is the address of the buffer.  */
+   variable_buffer is the address of the buffer.
+
+   For efficiency, it's guaranteed that the buffer will always have
+   VARIABLE_BUFFER_ZONE extra bytes allocated.  This allows you to add a few
+   extra chars without having to call a function.  Note you should never use
+   these bytes unless you're _sure_ you have room (you know when the buffer
+   length was last checked.  */
+
+#define VARIABLE_BUFFER_ZONE    5
 
 static unsigned int variable_buffer_length;
-static char *variable_buffer;
+char *variable_buffer;
 
 /* Subroutine of variable_expand and friends:
    The text to add is LENGTH chars starting at STRING to the variable_buffer.
@@ -44,7 +57,7 @@ variable_buffer_output (ptr, string, length)
 {
   register unsigned int newlen = length + (ptr - variable_buffer);
 
-  if (newlen > variable_buffer_length)
+  if ((newlen + VARIABLE_BUFFER_ZONE) > variable_buffer_length)
     {
       unsigned int offset = ptr - variable_buffer;
       variable_buffer_length = (newlen + 100 > 2 * variable_buffer_length
@@ -78,6 +91,8 @@ initialize_variable_output ()
 
 /* Recursively expand V.  The returned string is malloc'd.  */
 
+static char *allocated_variable_append PARAMS ((struct variable *v));
+
 char *
 recursively_expand (v)
      register struct variable *v;
@@ -85,20 +100,16 @@ recursively_expand (v)
   char *value;
 
   if (v->expanding)
-    {
-      /* Expanding V causes infinite recursion.  Lose.  */
-      if (reading_filename == 0)
-	fatal ("Recursive variable `%s' references itself (eventually)",
-	       v->name);
-      else
-	makefile_fatal
-	  (reading_filename, *reading_lineno_ptr, 
-	   "Recursive variable `%s' references itself (eventually)",
-	   v->name);
-    }
+    /* Expanding V causes infinite recursion.  Lose.  */
+    fatal (reading_file,
+           _("Recursive variable `%s' references itself (eventually)"),
+           v->name);
 
   v->expanding = 1;
-  value = allocated_variable_expand (v->value);
+  if (v->append)
+    value = allocated_variable_append (v);
+  else
+    value = allocated_variable_expand (v->value);
   v->expanding = 0;
 
   return value;
@@ -115,14 +126,8 @@ warn_undefined (name, length)
      unsigned int length;
 {
   if (warn_undefined_variables_flag)
-    {
-      static const char warnmsg[] = "warning: undefined variable `%.*s'";
-      if (reading_filename != 0)
-	makefile_error (reading_filename, *reading_lineno_ptr,
-			warnmsg, length, name);
-      else
-	error (warnmsg, length, name);
-    }
+    error (reading_file,
+           _("warning: undefined variable `%.*s'"), (int)length, name);
 }
 
 /* Expand a simple reference to variable NAME, which is LENGTH chars long.  */
@@ -136,36 +141,59 @@ reference_variable (o, name, length)
      char *name;
      unsigned int length;
 {
-  register struct variable *v = lookup_variable (name, length);
+  register struct variable *v;
+  char *value;
+
+  v = lookup_variable (name, length);
 
   if (v == 0)
     warn_undefined (name, length);
 
-  if (v != 0 && *v->value != '\0')
-    {
-      char *value = (v->recursive ? recursively_expand (v) : v->value);
-      o = variable_buffer_output (o, value, strlen (value));
-      if (v->recursive)
-	free (value);
-    }
+  if (v == 0 || *v->value == '\0')
+    return o;
+
+  value = (v->recursive ? recursively_expand (v) : v->value);
+
+  o = variable_buffer_output (o, value, strlen (value));
+
+  if (v->recursive)
+    free (value);
 
   return o;
 }
 
-/* Scan LINE for variable references and expansion-function calls.
-   Build in `variable_buffer' the result of expanding the references and calls.
-   Return the address of the resulting string, which is null-terminated
-   and is valid only until the next time this function is called.  */
+/* Scan STRING for variable references and expansion-function calls.  Only
+   LENGTH bytes of STRING are actually scanned.  If LENGTH is -1, scan until
+   a null byte is found.
+
+   Write the results to LINE, which must point into `variable_buffer'.  If
+   LINE is NULL, start at the beginning of the buffer.
+   Return a pointer to LINE, or to the beginning of the buffer if LINE is
+   NULL.  */
 
 char *
-variable_expand (line)
+variable_expand_string (line, string, length)
      register char *line;
+     char *string;
+     long length;
 {
   register struct variable *v;
   register char *p, *o, *p1;
+  char save_char = '\0';
+  unsigned int line_offset;
 
-  p = line;
-  o = initialize_variable_output ();
+  if (!line)
+    line = initialize_variable_output();
+
+  p = string;
+  o = line;
+  line_offset = line - variable_buffer;
+
+  if (length >= 0)
+    {
+      save_char = string[length];
+      string[length] = '\0';
+    }
 
   while (1)
     {
@@ -173,7 +201,7 @@ variable_expand (line)
          variable output buffer, and skip them.  Uninteresting chars end
 	 at the next $ or the end of the input.  */
 
-      p1 = index (p, '$');
+      p1 = strchr (p, '$');
 
       o = variable_buffer_output (o, p, p1 != 0 ? p1 - p : strlen (p) + 1);
 
@@ -197,9 +225,9 @@ variable_expand (line)
 	    char openparen = *p;
 	    char closeparen = (openparen == '(') ? ')' : '}';
 	    register char *beg = p + 1;
+	    int free_beg = 0;
 	    char *op, *begp;
-	    char *end;
-	    char *colon = 0;
+	    char *end, *colon;
 
 	    op = o;
 	    begp = p;
@@ -213,10 +241,12 @@ variable_expand (line)
 	    /* Is there a variable reference inside the parens or braces?
 	       If so, expand it before expanding the entire reference.  */
 
-	    p1 = index (beg, closeparen);
+	    end = strchr (beg, closeparen);
+	    if (end == 0)
+              /* Unterminated variable reference.  */
+              fatal (reading_file, _("unterminated variable reference"));
+	    p1 = lindex (beg, end, '$');
 	    if (p1 != 0)
-	      p1 = lindex (beg, p1, '$');
-	    if (p1 != 0 && lindex (beg, p1, ':') == 0)
 	      {
 		/* BEG now points past the opening paren or brace.
 		   Count parens or braces until it is matched.  */
@@ -227,133 +257,109 @@ variable_expand (line)
 		      ++count;
 		    else if (*p == closeparen && --count < 0)
 		      break;
-		    else if (colon == 0 && count == 0 && *p == ':')
-		      /* Record where we found a colon, which
-			 indicates a substitution reference.
-			 We want to expand the text before the
-			 reference only.  */
-		      colon = p;
 		  }
 		/* If COUNT is >= 0, there were unmatched opening parens
 		   or braces, so we go to the simple case of a variable name
 		   such as `$($(a)'.  */
 		if (count < 0)
 		  {
-		    char *name = expand_argument (beg, colon == 0 ? p : colon);
-		    unsigned int namelen = strlen (name);
-		    if (colon == 0)
-		      {
-			/* This is a simple reference to the expanded name.  */
-			o = reference_variable (o, name, namelen);
-			free (name);
-			break;
-		      }
-		    else
-		      {
-			/* This is a substitution reference to the expanded
-			   name.  We replace the pending text with a copy
-			   containing the expanded name in place of the
-			   original name, and then fall through to
-			   the normal substitution reference code below.  */
-			unsigned int restlen = strlen (colon) + 1;
-			beg = (char *) alloca (namelen + restlen);
-			bcopy (name, beg, namelen);
-			bcopy (colon, &beg[namelen], restlen);
-			/* Point COLON into the new copy.  */
-			colon = &beg[namelen];
-		      }
+		    beg = expand_argument (beg, p); /* Expand the name.  */
+		    free_beg = 1; /* Remember to free BEG when finished.  */
+		    end = strchr (beg, '\0');
 		  }
 	      }
+	    else
+	      /* Advance P to the end of this reference.  After we are
+                 finished expanding this one, P will be incremented to
+                 continue the scan.  */
+	      p = end;
 
 	    /* This is not a reference to a built-in function and
-	       it does not contain any variable references inside.
-	       There are several things it could be.  */
+	       any variable references inside are now expanded.
+	       Is the resultant text a substitution reference?  */
 
-	    if (colon == 0)
-	      colon = index (beg, ':');
-	    if (colon != 0 && lindex (beg, colon, closeparen) == 0)
+	    colon = lindex (beg, end, ':');
+	    if (colon != 0)
 	      {
-		/* This is a substitution reference: $(FOO:A=B).  */
-		int count;
+		/* This looks like a substitution reference: $(FOO:A=B).  */
 		char *subst_beg, *subst_end, *replace_beg, *replace_end;
 
-		v = lookup_variable (beg, colon - beg);
-		if (v == 0)
-		  warn_undefined (beg, colon - beg);
-
 		subst_beg = colon + 1;
-		count = 0;
-		for (p = subst_beg; *p != '\0'; ++p)
+		subst_end = strchr (subst_beg, '=');
+		if (subst_end == 0)
+		  /* There is no = in sight.  Punt on the substitution
+		     reference and treat this as a variable name containing
+		     a colon, in the code below.  */
+		  colon = 0;
+		else
 		  {
-		    if (*p == openparen)
-		      ++count;
-		    else if (*p == closeparen)
-		      --count;
-		    else if (*p == '=' && count <= 0)
-		      break;
+		    replace_beg = subst_end + 1;
+		    replace_end = end;
+
+		    /* Extract the variable name before the colon
+		       and look up that variable.  */
+		    v = lookup_variable (beg, colon - beg);
+		    if (v == 0)
+		      warn_undefined (beg, colon - beg);
+
+		    if (v != 0 && *v->value != '\0')
+		      {
+			char *value = (v->recursive ? recursively_expand (v)
+				       : v->value);
+			char *pattern, *percent;
+			if (free_beg)
+			  {
+			    *subst_end = '\0';
+			    pattern = subst_beg;
+			  }
+			else
+			  {
+			    pattern = (char *) alloca (subst_end - subst_beg
+						       + 1);
+			    bcopy (subst_beg, pattern, subst_end - subst_beg);
+			    pattern[subst_end - subst_beg] = '\0';
+			  }
+			percent = find_percent (pattern);
+			if (percent != 0)
+			  {
+			    char *replace;
+			    if (free_beg)
+			      {
+				*replace_end = '\0';
+				replace = replace_beg;
+			      }
+			    else
+			      {
+				replace = (char *) alloca (replace_end
+							   - replace_beg
+							   + 1);
+				bcopy (replace_beg, replace,
+				       replace_end - replace_beg);
+				replace[replace_end - replace_beg] = '\0';
+			      }
+
+			    o = patsubst_expand (o, value, pattern, replace,
+						 percent, (char *) 0);
+			  }
+			else
+			  o = subst_expand (o, value,
+					    pattern, replace_beg,
+					    strlen (pattern),
+					    end - replace_beg,
+					    0, 1);
+			if (v->recursive)
+			  free (value);
+		      }
 		  }
-		if (count > 0)
-		  /* There were unmatched opening parens.  */
-		  return initialize_variable_output ();
-		subst_end = p;
-
-		replace_beg = p + 1;
-		count = 0;
-		for (p = replace_beg; *p != '\0'; ++p)
-		  {
-		    if (*p == openparen)
-		      ++count;
-		    else if (*p == closeparen && --count < 0)
-		      break;
-		  }
-		if (count > 0)
-		  /* There were unmatched opening parens.  */
-		  return initialize_variable_output ();
-		end = p;
-		replace_end = p;
-
-		p = expand_argument (subst_beg, subst_end);
-		p1 = expand_argument (replace_beg, replace_end);
-
-		if (v != 0 && *v->value != '\0')
-		  {
-		    char *value = (v->recursive ? recursively_expand (v)
-				   : v->value);
-		    char *percent = find_percent (p);
-		    if (percent != 0)
-		      o = patsubst_expand (o, value, p, p1,
-					   percent, (char *) 0);
-		    else
-		      o = subst_expand (o, value,
-					p, p1, strlen (p), strlen (p1),
-					0, 1);
-		    if (v->recursive)
-		      free (value);
-		  }
-
-		free (p);
-		free (p1);
 	      }
 
-	    /* No, this must be an ordinary variable reference.  */
-	    else
-	      {
-		/* Look up the value of the variable.  */
-		end = index (beg, closeparen);
-		if (end == 0)
-		  {
-		    /* Unterminated variable reference.  */
-		    if (reading_filename != 0)
-		      makefile_fatal (reading_filename, *reading_lineno_ptr,
-				      "unterminated variable reference");
-		    else
-		      fatal ("unterminated variable reference");
-		  }
+	    if (colon == 0)
+	      /* This is an ordinary variable reference.
+		 Look up the value of the variable.  */
 		o = reference_variable (o, beg, end - beg);
-	      }
 
-	    /* Advance p past the variable reference to resume scan.  */
-	    p = end;
+	  if (free_beg)
+	    free (beg);
 	  }
 	  break;
 
@@ -381,7 +387,7 @@ variable_expand (line)
 	  }
 
 	  break;
-	}      
+	}
 
       if (*p == '\0')
 	break;
@@ -389,8 +395,23 @@ variable_expand (line)
 	++p;
     }
 
-  (void) variable_buffer_output (o, "", 1);
-  return initialize_variable_output ();
+  if (save_char)
+    string[length] = save_char;
+
+  (void)variable_buffer_output (o, "", 1);
+  return (variable_buffer + line_offset);
+}
+
+/* Scan LINE for variable references and expansion-function calls.
+   Build in `variable_buffer' the result of expanding the references and calls.
+   Return the address of the resulting string, which is null-terminated
+   and is valid only until the next time this function is called.  */
+
+char *
+variable_expand (line)
+     char *line;
+{
+  return variable_expand_string(NULL, line, (long)-1);
 }
 
 /* Expand an argument for an expansion function.
@@ -405,7 +426,7 @@ expand_argument (str, end)
 {
   char *tmp;
 
-  if (*end == '\0')
+  if (!end || *end == '\0')
     tmp = str;
   else
     {
@@ -420,7 +441,7 @@ expand_argument (str, end)
 /* Expand LINE for FILE.  Error messages refer to the file and line where
    FILE's commands were found.  Expansion uses FILE's variable set list.  */
 
-char *
+static char *
 variable_expand_for_file (line, file)
      char *line;
      register struct file *file;
@@ -433,16 +454,65 @@ variable_expand_for_file (line, file)
 
   save = current_variable_set_list;
   current_variable_set_list = file->variables;
-  reading_filename = file->cmds->filename;
-  reading_lineno_ptr = &file->cmds->lineno;
+  if (file->cmds && file->cmds->fileinfo.filenm)
+    reading_file = &file->cmds->fileinfo;
+  else
+    reading_file = 0;
   result = variable_expand (line);
   current_variable_set_list = save;
-  reading_filename = 0;
-  reading_lineno_ptr = 0;
+  reading_file = 0;
 
   return result;
 }
 
+/* Like allocated_variable_expand, but we first expand this variable in the
+    context of the next variable set, then we append the expanded value.  */
+
+static char *
+allocated_variable_append (v)
+     struct variable *v;
+{
+  struct variable_set_list *save;
+  int len = strlen (v->name);
+  char *var = alloca (len + 4);
+  char *value;
+
+  char *obuf = variable_buffer;
+  unsigned int olen = variable_buffer_length;
+
+  variable_buffer = 0;
+
+  assert(current_variable_set_list->next != 0);
+  save = current_variable_set_list;
+  current_variable_set_list = current_variable_set_list->next;
+
+  var[0] = '$';
+  var[1] = '(';
+  strcpy (&var[2], v->name);
+  var[len+2] = ')';
+  var[len+3] = '\0';
+
+  value = variable_expand_for_file (var, 0);
+
+  current_variable_set_list = save;
+
+  value += strlen (value);
+  value = variable_buffer_output (value, " ", 1);
+  value = variable_expand_string (value, v->value, (long)-1);
+
+  value = variable_buffer;
+
+#if 0
+  /* Waste a little memory and save time.  */
+  value = xrealloc (value, strlen (value))
+#endif
+
+  variable_buffer = obuf;
+  variable_buffer_length = olen;
+
+  return value;
+}
+
 /* Like variable_expand_for_file, but the returned string is malloc'd.
    This function is called a lot.  It wants to be efficient.  */
 
